# Autogenerated by Astropy's setup.py on 2017-09-08 16:26:42.927375
from __future__ import unicode_literals
import datetime

version = "2.0.2"
githash = "93f7e758ee3113f83a150cc98ce52c1eb894b16e"


major = 2
minor = 0
bugfix = 2

release = True
timestamp = datetime.datetime(2017, 9, 8, 16, 26, 42, 927375)
debug = False

try:
    from ._compiler import compiler
except ImportError:
    compiler = "unknown"

try:
    from .cython_version import cython_version
except ImportError:
    cython_version = "unknown"
